use std::collections::HashMap;

use crate::{OpCode, LispValue};

pub struct LispState {
	table: HashMap<String, LispValue>,
}

fn add_function<T: Into<String> + Clone>(table: &mut HashMap<String, LispValue>, name: T, func: fn(Vec<LispValue>) -> Vec<LispValue>) {
	table.insert(name.clone().into(), LispValue::RustFunction(name.into(), func));
}

impl LispState {
	pub fn new() -> LispState {
		let mut table = HashMap::new();

		add_function(&mut table, "print", |x| {
			let mut strings = Vec::new();
			for val in x {
				strings.push(val.to_string());
			}

			let str = strings.join(" ");
			println!("{}", str);

			vec![LispValue::Nil]
		});

		add_function(&mut table, "concat", |x| {
			let mut strings = Vec::new();
			for val in x {
				strings.push(val.to_string());
			}

			let str = strings.concat();
			vec![LispValue::String(str)]
		});

		add_function(&mut table, "add", |x| {
			if x.len() < 1 {
				return vec![LispValue::Nil]
			} else if x.len() == 1 {
				return vec![x[0].clone()];
			}

			let mut t = "int";
			for v in &x {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return vec![LispValue::Nil]
				}
			}

			match t {
				"int" => {
					let mut sum = 0;
					for v in &x {
						if let LispValue::Integer(n) = v {
							sum += n;
						} else {
							unreachable!()
						}
					}
					return vec![LispValue::Integer(sum)];
				},
				"float" => {
					let mut sum = 0.0 as f32;
					for v in &x {
						if let LispValue::Float(n) = v {
							sum += n;
						} else if let LispValue::Integer(n) = v{
							sum += *n as f32;
						} else {
							unreachable!()
						}
					}
					return vec![LispValue::Float(sum)];
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "sub", |x| {
			if x.len() < 1 {
				return vec![LispValue::Nil]
			} else if x.len() == 1 {
				return vec![x[0].clone()];
			}

			let mut t = "int";
			for v in &x {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return vec![LispValue::Nil]
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = x[0] {
						let mut sum = n;
						for i in 1..x.len() {
							if let LispValue::Integer(n) = &x[i] {
								sum -= n;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Integer(sum)];
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = x[0] {
						let mut sum = n;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(sum)];
					} else if let LispValue::Integer(n) = x[0] {
						let mut sum = n as f32;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(sum)];
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "mul", |x| {
			if x.len() < 1 {
				return vec![LispValue::Nil]
			} else if x.len() == 1 {
				return vec![x[0].clone()];
			}

			let mut t = "int";
			for v in &x {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return vec![LispValue::Nil]
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = x[0] {
						let mut sum = n;
						for i in 1..x.len() {
							if let LispValue::Integer(n) = &x[i] {
								sum *= n;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Integer(sum)];
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = x[0] {
						let mut product = n;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(product)];
					} else if let LispValue::Integer(n) = x[0] {
						let mut product = n as f32;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(product)];
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "div", |x| {
			if x.len() < 1 {
				return vec![LispValue::Nil]
			} else if x.len() == 1 {
				return vec![x[0].clone()];
			}

			let mut t = "int";
			for v in &x {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return vec![LispValue::Nil]
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = x[0] {
						let mut sum = n;
						for i in 1..x.len() {
							if let LispValue::Integer(n) = &x[i] {
								sum /= n;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Integer(sum)];
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = x[0] {
						let mut product = n;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(product)];
					} else if let LispValue::Integer(n) = x[0] {
						let mut product = n as f32;
						for i in 1..x.len() {
							if let LispValue::Float(n) = &x[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &x[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return vec![LispValue::Float(product)];
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		LispState {
			table
		}
	}

	fn do_call(&self, opcode: &OpCode) -> Vec<LispValue> {
		if let OpCode::Call(func, args) = opcode {
			let f = self.table.get(func).unwrap();

			let mut args: Vec<LispValue> = args.clone();
			for i in 0..args.len() {
				let arg = &args[i];
				if let LispValue::Eval(OpCode::Exp(opcodes)) = arg {
					if opcodes.len() < 1 {
						args[i] = LispValue::Nil;
						continue;
					}

					let call_opcode = opcodes.last().unwrap();
					let mut res = self.do_call(call_opcode);
					args.remove(i);
					for j in 0..res.len() {
						args.insert(i + j, res.remove(0));
					}
				}
			}
			if let LispValue::RustFunction(_, f) = f {
				return f(args.to_vec());
			} else {
				todo!();
			}

		} else {
			panic!("Bad OpCode: expected OpCode::Call")
		}
	}

	pub fn execute(&self, instructions: Vec<OpCode>) {
		for op in instructions {
			match op {
				OpCode::Call(func, args) => {
					self.do_call(&OpCode::Call(func, args));
				},
				OpCode::Exp(ins) =>  {
					self.execute(ins);
				}
			}
		}
	}
}
